import 'dart:convert';
import 'dart:async';
import 'dart:math';

import 'package:web_socket_channel/io.dart';
// import 'package:web_socket_channel/status.dart' as status;
// import 'package:flutter/foundation.dart';

var random = Random.secure();

// ----------------------------------------------------------------------------
// Exported classes
// ----------------------------------------------------------------------------

class VocGateway {
  /// The WebSocket "open" channel
  IOWebSocketChannel _channel;
  List<VocGatewayRequest> _requests = [];
  String publicKey;

  // /// List of callbacks to invoke when an unrelated message is received.
  // ObserverList<Function> _listeners = new ObserverList<Function>();

  VocGateway(String uri, {String publicKey}) {
    this.publicKey = publicKey;
    connect(uri);
  }

  /// Initialization the WebSockets connection with the server
  connect(String uri) {
    if (uri == null || uri == "") throw "Invalid Gateway URI";

    disconnect();
    _channel = new IOWebSocketChannel.connect(uri);
    _channel.stream.listen(_gotMessage);
  }

  /// Disconnects the WebSocket communication
  disconnect() {
    if (_channel != null) {
      if (_channel.sink != null) {
        _channel.sink.close();
      }
    }
  }

  /// Perform a request to the Vocdoni Gateway and wait for a response to
  /// arrive within a given timeout
  Future<dynamic> request(Map<String, dynamic> requestBody,
      {int timeout = 40, String privateKey}) {
    if (_channel == null || _channel.sink == null)
      throw "You are not connected to a gateway";

    var values = List<int>.generate(16, (i) => random.nextInt(256));
    final String id = base64Url.encode(values);

    Completer comp = Completer();
    VocGatewayRequest req = VocGatewayRequest(id, comp);
    _requests.add(req);

    Future.delayed(Duration(seconds: timeout)).then((_) {
      if (comp.isCompleted) return;
      comp.completeError("The request timed out");
    });

    Map<String, dynamic> content;

    // Sign if needed
    if (privateKey == null || privateKey == "") {
      content = {"id": id, "request": requestBody};
    } else {
      final String signature = "";
      content = {"id": id, "request": requestBody, "signature": signature};
    }

    _channel.sink.add(jsonEncode(content));

    return comp.future;
  }

  /// Try to match the incoming message to a previous request
  _gotMessage(message) {
    var incomingMessage;

    try {
      incomingMessage = jsonDecode(message);
    } catch (err) {
      return print("ERR: Received a non-JSON message");
    }

    if (incomingMessage == null) {
      return print("ERR: Received an empty message");
    } else if (!(incomingMessage is Map)) {
      return print("ERR: Received an invalid message");
    } else if (incomingMessage["id"] == null) {
      return _handleUnrelatedMessage(incomingMessage);
    }

    final req = _requests.firstWhere((req) => req.id == incomingMessage["id"]);
    if (req == null) {
      // NO ORIGINARY REQUEST
      return print("ERR: Received a message for a non-existing request: " +
          incomingMessage["id"]);
    } else if (req.completer.isCompleted) {
      // ALREADY HANDLED
      return print("Got a response for an already completed request");
    }

    _requests.remove(req);

    if (incomingMessage["error"] != null) {
      final error = jsonDecode(incomingMessage["error"]);

      // ERROR RESPONSE
      if (!(error is Map)) {
        return req.completer
            .completeError("Received a non-object 'error' response");
      } else if (incomingMessage["id"] != error["requestId"]) {
        return req.completer.completeError(
            "The signed requestId does not match the expected one");
      } else if (!_isValidResponseSignature(
          incomingMessage["signature"], error)) {
        return req.completer
            .completeError("Received an invalid response signature");
      } else if (!error["message"] is String) {
        return req.completer.completeError("Received an empty error message");
      }
      return req.completer.completeError(error["message"]);
    }

    if (incomingMessage["response"] == null) {
      return req.completer.completeError("Received an empty response");
    }
    final response = jsonDecode(incomingMessage["response"]);

    if (!(response is Map)) {
      return req.completer.completeError("Received an invalid response");
    }

    // CHECK RESPONSE
    if (!_isValidResponseSignature(incomingMessage["signature"], response)) {
      return req.completer
          .completeError("Received an invalid response signature");
    }

    // DONE
    req.completer.complete(response);
  }

  // /// Adds a callback to be invoked in case of incoming notification
  // addListener(Function callback) {
  //   _listeners.add(callback);
  // }

  // removeListener(Function callback) {
  //   _listeners.remove(callback);
  // }

  _handleUnrelatedMessage(Map incomingMessage) {
    print("GOT AN UNRELATED MESSAGE: " + incomingMessage.toString());
    // _listeners.forEach((Function callback) {
    //   callback(message);
    // });
  }

  String _signMessagePayload(Map<String, dynamic> payload, String privateKey) {
    // Ensure alphabetically ordered key names
    payload = sortMapFields(payload);
    final data = jsonEncode(payload);

    // TODO: IMPLEMENT SIGNING
    // https://goethereumbook.org/signature-generate/

    return "0x1234";
  }

  bool _isValidResponseSignature(
      String signature, Map<String, dynamic> payload) {
    if (publicKey == null || publicKey == "") return true;

    // Reorder fields alphabetically
    payload = sortMapFields(payload);

    // TODO: IMPLEMENT SIGNING
    // https://goethereumbook.org/signature-generate/

    // TODO: VALIDATE THE RESPONSE SIGNATURE

    return true;
  }
}

/// Client class to wrap calls to Ethereum Smart Contracts using a (Vocdoni) Gateway
class Web3Gateway {
  // TODO:
  // https://github.com/simolus3/web3dart/blob/development/example/contracts.dart
}

// ----------------------------------------------------------------------------
// Internal classes
// ----------------------------------------------------------------------------

class VocGatewayRequest {
  String id;
  Completer completer;

  VocGatewayRequest(this.id, this.completer);
}

// ----------------------------------------------------------------------------
// Helper functions
// ----------------------------------------------------------------------------

/// Signatures need to be computed over objects that can be 100% reproduceable. 
/// Since the ordering is not guaranteed, this function returns a recursively 
/// ordered map
Map<String, dynamic> sortMapFields(Map<String, dynamic> data) {
  List<String> keys = [];
  Map<String, dynamic> result = Map<String, dynamic>();

  data.forEach((k, v) {
    keys.add(k);
  });
  keys.sort((String a, String b) => a.compareTo(b));
  keys.forEach((k) {
    if (data[k] is Map) {
      result[k] = sortMapFields(data[k]);
    } else {
      result[k] = data[k];
    }
  });
  return result;
}
